package geometry.part10;

import geometry.part1.Line;
import geometry.part2.LineSegment;
import geometry.part6.ConvexPolygon;
import geometry.part6.GeomUtils;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class PolygonIntersectionCalculator {
    // 走査線上での辺の相対位置
    enum EdgePosition {
        LEFT1, RIGHT1, LEFT2, RIGHT2
    }

    // 入力となる凸多角形を左側と右側に分解して作られる辺連結
    static class Edge {
        LineSegment segment; // 辺の線分
        Point2D startPoint; // 辺の始点
        Point2D endPoint; // 辺の終点
        Edge next; // 次の辺

        boolean intersects(Edge other) {
            return segment.intersects(other.segment);
        }

        // 他の辺との交点を求める
        Point2D getIntersectionPoint(Edge other) {
            return segment.toLine()
                    .getIntersectionPoint(other.segment.toLine());
        }

        // 走査線との交点のx座標を求める
        double getIntersectionX(Line sweepLine) {
            Point2D p = segment.getIntersectionPoint(sweepLine);
            return p != null ? p.getX() : startPoint.getX();
        }
    }

    // 走査線と交差する4つの辺を格納するステータス
    static class Status {
        Edge left1;
        Edge right1;
        Edge left2;
        Edge right2;
    }

    // イベントの処理結果
    static class StepResult {
        double nextSweepY; // 次の走査線位置として設定すべきy座標
        boolean mustContinue; // これ以降も処理を継続する必要があるかどうか

        public StepResult(double nextSweepY, boolean mustContinue) {
            this.nextSweepY = nextSweepY;
            this.mustContinue = mustContinue;
        }
    }

    // polygon1とpolygon2の交差領域となる凸多角形を求める
    public ConvexPolygon execute(ConvexPolygon polygon1, ConvexPolygon polygon2) {
        Status status = new Status();
        List<Edge> leftResult = new ArrayList<Edge>(); // 左側の計算結果
        List<Edge> rightResult = new ArrayList<Edge>(); // 右側の計算結果

        // 最初のイベントを処理
        StepResult result = firstPass(polygon1, polygon2, status, leftResult,
                rightResult);
        while (result.mustContinue) {
            // 2番目以降のイベントを処理
            result = secondPass(status, leftResult, rightResult);
        }

        // 左側と右側の計算結果を統合するリスト
        List<Edge> totalResult = new ArrayList<Edge>();
        totalResult.addAll(leftResult);
        // 左側と右側の結果は辺の巡回方向が逆になっているので
        // 右側の結果を反転して連結
        Collections.reverse(rightResult);
        totalResult.addAll(rightResult);

        List<Point2D> resultPoints = new ArrayList<Point2D>();
        Point2D lastPoint = null;
        int totalSize = totalResult.size();
        for (int i = 0; i < totalSize; i++) {
            Edge e1 = totalResult.get(i);
            Edge e2 = totalResult.get((i + 1) % totalSize);
            Point2D p = e1.getIntersectionPoint(e2);
            if (p != null && !p.equals(lastPoint)) {
                resultPoints.add(p);
                lastPoint = p;
            }
        }

        // 3点以上が求められたら交差結果を凸多角形として返す
        if (resultPoints.size() >= 3) {
            return new ConvexPolygon(resultPoints);
        } else {
            return null;
        }
    }

    // 初回のイベント処理。凸多角形の左右の辺連結を作成し、走査線の初期位置を設定する
    // （デモプログラムからも使用するため、メソッドをデフォルトスコープで宣言している）
    StepResult firstPass(ConvexPolygon polygon1, ConvexPolygon polygon2,
            Status status, List<Edge> leftResult, List<Edge> rightResult) {
        // 凸多角形の左右の辺連結を作成
        Edge left1 = createEdgeChain(polygon1, true);
        Edge right1 = createEdgeChain(polygon1, false);
        Edge left2 = createEdgeChain(polygon2, true);
        Edge right2 = createEdgeChain(polygon2, false);

        // Edgeチェインの最上点の座標を取得
        double topX1 = left1.startPoint.getX();
        double topY1 = left1.startPoint.getY();
        double topX2 = left2.startPoint.getX();
        double topY2 = left2.startPoint.getY();

        // 2つの凸多角形の最上点のうち、より下にある方のy座標を走査線の初期位置とする
        double sweepY = Math.max(topY1, topY2);
        // 走査線を作成
        Line sweepLine = Line.fromPoints(0, sweepY, 1, sweepY);

        // Edgeチェインの中から、初めに走査線と交わるEdgeを見つけてステータスに設定
        status.left1 = findInitialEdge(left1, sweepLine);
        status.right1 = findInitialEdge(right1, sweepLine);
        status.left2 = findInitialEdge(left2, sweepLine);
        status.right2 = findInitialEdge(right2, sweepLine);

        // いずれかのEdgeが見つからなければ、交差部分は存在しないので終了
        if (status.left1 == null || status.right1 == null
                || status.left2 == null || status.right2 == null) {
            return new StepResult(sweepY, false);
        }

        // 初回のイベント処理。2つの最上点の位置関係によって、必要な処理とその順番が変わる
        if (topY1 > topY2) {
            if (topX1 > topX2) {
                process(status, EdgePosition.LEFT1, sweepLine, leftResult,
                        rightResult);
                process(status, EdgePosition.RIGHT1, sweepLine, leftResult,
                        rightResult);
            } else {
                process(status, EdgePosition.RIGHT1, sweepLine, leftResult,
                        rightResult);
                process(status, EdgePosition.LEFT1, sweepLine, leftResult,
                        rightResult);
            }
        } else {
            if (topX1 > topX2) {
                process(status, EdgePosition.RIGHT2, sweepLine, leftResult,
                        rightResult);
                process(status, EdgePosition.LEFT2, sweepLine, leftResult,
                        rightResult);
            } else {
                process(status, EdgePosition.LEFT2, sweepLine, leftResult,
                        rightResult);
                process(status, EdgePosition.RIGHT2, sweepLine, leftResult,
                        rightResult);
            }
        }

        return new StepResult(sweepY, true);
    }

    // 2回目以降のイベント処理
    // デモプログラムからも使用するためスコープをパッケージプライベートとしている
    StepResult secondPass(Status status, List<Edge> leftResult,
            List<Edge> rightResult) {
        // 次に処理すべき辺を選択
        Edge next = pickNextEdge(status);
        if (next == null) { // 見つからなければ終了
            return new StepResult(Double.NaN, false);
        }

        // 選択された辺の始点を、走査線の次の位置とする
        double nextSweepY = next.startPoint.getY();
        // 次の走査線を作成
        Line sweepLine = Line.fromPoints(0, nextSweepY, 1, nextSweepY);

        // 選択された辺が、ステータスの中でleft1/right1/left2/right2のうち
        // どれと対応するかを特定する
        EdgePosition pos = null;
        if (next == status.left1.next) {
            pos = EdgePosition.LEFT1;
            status.left1 = next;
        } else if (next == status.right1.next) {
            pos = EdgePosition.RIGHT1;
            status.right1 = next;
        } else if (next == status.left2.next) {
            pos = EdgePosition.LEFT2;
            status.left2 = next;
        } else if (next == status.right2.next) {
            pos = EdgePosition.RIGHT2;
            status.right2 = next;
        }

        // イベント処理
        process(status, pos, sweepLine, leftResult, rightResult);
        return new StepResult(nextSweepY, true);
    }

    // ステータス中でposの位置にある辺の始点をイベント点として処理する
    // 交差部分の辺を左右別に求め、leftResultとrightResultに追加する
    private void process(Status status, EdgePosition pos, Line sweepLine,
            List<Edge> leftResult, List<Edge> rightResult) {
        switch (pos) {
        case LEFT1:
            processLeft(status.left1, status.right1, status.left2,
                    status.right2, sweepLine, leftResult, rightResult);
            break;
        case RIGHT1:
            processRight(status.left1, status.right1, status.left2,
                    status.right2, sweepLine, leftResult, rightResult);
            break;
        case LEFT2:
            processLeft(status.left2, status.right2, status.left1,
                    status.right1, sweepLine, leftResult, rightResult);
            break;
        case RIGHT2:
            processRight(status.left2, status.right2, status.left1,
                    status.right1, sweepLine, leftResult, rightResult);
            break;
        default:
            throw new IllegalStateException();
        }
    }

    // left1の始点をイベント点として処理する
    // 処理の内容はprocessRight()と完全に対称
    private void processLeft(Edge left1, Edge right1, Edge left2, Edge right2,
            Line sweepLine, List<Edge> leftResult, List<Edge> rightResult) {
        double l1 = left1.getIntersectionX(sweepLine);
        double l2 = left2.getIntersectionX(sweepLine);
        double r2 = right2.getIntersectionX(sweepLine);
        // left1がleft2とright2の内部から始まる場合
        if (l1 > l2 && l1 < r2) {
            leftResult.add(left1); // left1は交差凸多角形の一部
        }
        // left1がright2と交わり、right2よりも右から始まる場合
        if (left1.intersects(right2) && l1 >= r2) {
            // left1, right2はともに交差凸多角形の一部、かつ
            // 必ず上端となるため結果の先頭位置に追加
            leftResult.add(0, left1);
            rightResult.add(0, right2);
        }
        // left1がleft2と交わる場合
        if (left1.intersects(left2)) {
            if (l1 > l2) { // left1がleft2よりも右から始まるなら
                leftResult.add(left2); // left2は交差凸多角形の一部
            } else { // そうでないなら
                leftResult.add(left1); // left1は交差凸多角形の一部
            }
        }
    }

    // right1の始点をイベント点として処理する
    // 処理の内容はprocessLeft()と完全に対称
    private void processRight(Edge left1, Edge right1, Edge left2, Edge right2,
            Line sweepLine, List<Edge> leftResult, List<Edge> rightResult) {
        double r1 = right1.getIntersectionX(sweepLine);
        double l2 = left2.getIntersectionX(sweepLine);
        double r2 = right2.getIntersectionX(sweepLine);
        // right1がleft2とright2の内部から始まる場合
        if (r1 > l2 && r1 < r2) {
            rightResult.add(right1); // right1は交差凸多角形の一部
        }
        // right1がleft2と交わり、left2よりも左から始まる場合
        if (right1.intersects(left2) && r1 <= l2) {
            // right1, left2はともに交差凸多角形の一部、かつ
            // 必ず上端となるため結果の先頭位置に追加
            rightResult.add(0, right1);
            leftResult.add(0, left2);
        }
        // right1がright2と交わる場合
        if (right1.intersects(right2)) {
            if (r1 < r2) { // rightがright2よりも左から始まるなら
                rightResult.add(right2); // right2は交差凸多角形の一部
            } else { // そうでないなら
                rightResult.add(right1); // right1は交差凸多角形の一部
            }
        }
    }

    // 辺連結の中から、初めに走査線と交わる辺を見つける
    private Edge findInitialEdge(Edge edge, Line sweepLine) {
        for (Edge e = edge; e != null; e = e.next) {
            if (e.segment.intersects(sweepLine)) {
                return e;
            }
        }
        return null;
    }

    // 凸多角形の「左側」または「右側」の、上から下に向かう辺連結を作成する
    private Edge createEdgeChain(ConvexPolygon polygon, boolean left) {
        double minY = Double.POSITIVE_INFINITY;
        double maxY = Double.NEGATIVE_INFINITY;
        int minYIndex = -1;
        int maxYIndex = -1;
        int size = polygon.getEdgeCount();
        // 凸多角形の頂点のうち、最上点と最下点の位置を特定
        for (int i = 0; i < size; i++) {
            Point2D v = polygon.getVertex(i);
            double y = v.getY();
            if (y < minY) {
                minY = y;
                minYIndex = i;
            }
            if (y > maxY) {
                maxY = y;
                maxYIndex = i;
            }
        }

        Edge firstEdge = null;
        Edge lastEdge = null;

        boolean ccw = GeomUtils.ccw(polygon.getVertex(0), polygon.getVertex(1),
                polygon.getVertex(2)) < 0;
        // 凸多角形の辺配列を前進しながら辿るか、後退しながら辿るかを決定
        boolean forward = (ccw && left) || (!ccw && !left);
        int nextIndex;
        // 最上点の位置から開始し、最下点に到達するまで続ける
        for (int i = minYIndex; i != maxYIndex; i = nextIndex) {
            Edge edge = new Edge(); // 辺を作成
            if (forward) { // 前進する場合
                nextIndex = (i + 1) % size;
                edge.segment = polygon.getEdge(i);
            } else { // 後退する場合
                nextIndex = (i + size - 1) % size;
                edge.segment = polygon.getEdge(nextIndex);
            }
            // 辺の始点と終点を代入
            edge.startPoint = polygon.getVertex(i);
            edge.endPoint = polygon.getVertex(nextIndex);
            // 初めて辺が作成された場合
            if (firstEdge == null) {
                firstEdge = edge; // 先頭の辺として記憶
            } else { // そうでない場合
                // 直前の辺に新しい辺を連結
                lastEdge.next = edge;
            }
            lastEdge = edge;
        }

        return firstEdge;
    }

    // ステータスの中から終点のy座標がもっとも上にある辺を探し、次に処理すべき辺として返す
    private Edge pickNextEdge(Status status) {
        Edge result = chooseEdgeWithUpperEndY(status.left1, status.right1);
        result = chooseEdgeWithUpperEndY(result, status.left2);
        result = chooseEdgeWithUpperEndY(result, status.right2);
        return result.next;
    }

    // e1とe2のうち、終点のy座標がより上にあるものを選択する
    private Edge chooseEdgeWithUpperEndY(Edge e1, Edge e2) {
        double y1 = e1 != null ? e1.endPoint.getY() : Double.POSITIVE_INFINITY;
        double y2 = e2 != null ? e2.endPoint.getY() : Double.POSITIVE_INFINITY;
        if (y1 == y2) {
            if (e1 != null && e1.next != null) {
                return e1;
            } else {
                return e2;
            }
        } else if (y1 < y2) {
            return e1;
        } else {
            return e2;
        }
    }
}