package geometry.part6;

import geometry.part2.LineSegment;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

public class ConvexPolygon {
    private List<Point2D> vertices; // 頂点
    private List<LineSegment> edges; // 辺

    // verticesには凸多角形の頂点座標が順番に格納してあるものとする
    public ConvexPolygon(List<Point2D> vertices) {
        int size = vertices.size();
        if (size < 3) { // 角数が3未満の場合はエラー
            throw new IllegalArgumentException();
        }
        this.vertices = vertices;
        edges = new ArrayList<LineSegment>();

        // 基準となるCCW値を計算
        double ccw0 = GeomUtils.ccw(vertices.get(0), vertices.get(1), vertices
                .get(2));
        if (ccw0 == 0) { // ゼロの場合はエラー
            throw new IllegalArgumentException("Polygon is not convex.");
        }
        for (int i = 1; i < size; i++) {
            Point2D v1 = vertices.get(i); // i番目の頂点
            Point2D v2 = vertices.get((i + 1) % size); // v1の次の頂点
            Point2D v3 = vertices.get((i + 2) % size); // v2の次の頂点
            double ccw = GeomUtils.ccw(v1, v2, v3); // CCW値を計算
            if (ccw0 * ccw <= 0) { // 基準値と符号が異なる、またはゼロの場合はエラー
                throw new IllegalArgumentException("Polygon is not convex.");
            }
        }

        for (int i = 0; i < size; i++) {
            Point2D v1 = vertices.get(i); // i番目の頂点
            Point2D v2 = vertices.get((i + 1) % size); // v1の次の頂点
            // 2つの頂点から辺の線分を作成して登録
            edges.add(new LineSegment(v1.getX(), v1.getY(), v2.getX(), v2
                    .getY()));
        }
    }

    // 指定位置の頂点を取得
    public Point2D getVertex(int index) {
        return vertices.get(index);
    }

    // 指定位置の辺を取得
    public LineSegment getEdge(int index) {
        return edges.get(index);
    }

    // 辺の数（＝角数）を取得
    public int getEdgeCount() {
        return edges.size();
    }

    public boolean contains(double x, double y) {
        // 多角形のy座標範囲を求める
        double minY = Double.POSITIVE_INFINITY;
        double maxY = Double.NEGATIVE_INFINITY;
        for (Point2D v : vertices) {
            minY = Math.min(minY, v.getY());
            maxY = Math.max(maxY, v.getY());
        }
        // yが最小値-最大値の範囲外の場合はfalseを返す
        if (y <= minY || y >= maxY) {
            return false;
        }
        // 与えられた座標を始点とし、右方向に十分長く延びる擬似的な半直線を作成
        LineSegment halfLine = new LineSegment(x, y, x + 10000000, y);
        int count = 0;
        for (LineSegment edge : edges) {
            // 半直線が辺の終点とちょうど重なる場合、次の辺の始点とも交差が検出され、
            // 二重にカウントされてしまうため、カウントをスキップする
            if (edge.y2 == y) {
                continue;
            }
            if (edge.intersects(halfLine)) {
                count++;
            }
        }
        // 交差回数が奇数の場合は点を包含
        return count % 2 == 1;
    }

    public boolean contains(Point2D p) {
        return contains(p.getX(), p.getY());
    }

    public double getArea() {
        double crossSum = 0; // 外積の合計
        int size = vertices.size();
        // 頂点を巡回
        for (int i = 0; i < size; i++) {
            Point2D v1 = vertices.get(i);
            Point2D v2 = vertices.get((i + 1) % size);
            // 外積を計算
            double cross = GeomUtils.cross(v1, v2);
            crossSum += cross; // 外積を加算
        }
        return Math.abs(crossSum / 2.0);
    }
}
